// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 13:33 BRST
// Last edited on December 28 of 2017, at 21:11 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/multiboot.h>
#include <chicago/string.h>
#include <chicago/console.h>
#include <chicago/arch/gdt.h>
#include <chicago/arch/idt.h>
#include <chicago/timer.h>
#include <chicago/keyboard.h>
#include <chicago/arch/paging.h>
#include <chicago/alloc.h>
#include <chicago/task.h>

Void ArchInit(PMultibootHeader mbheader)
{
	/* Init GDT */
	
	GDTInit();
	Message(MESSAGE_LEVEL_INFO, "GDT initialized\n");
	
	/* Init IDT */
	
	IDTInit();
	Message(MESSAGE_LEVEL_INFO, "IDT initialized\n");
	
	/* Init timer */
	
	TimerInit(1000);
	Message(MESSAGE_LEVEL_INFO, "Timer initialized\n");
	
	/* Init keyboard */
	
	KeyboardInit();
	Message(MESSAGE_LEVEL_INFO, "Keyboard initialized\n");
	
	/* Init paging */
	
	PagingInit(mbheader);
	Message(MESSAGE_LEVEL_INFO, "Paging initialized\n");
	
	/* Init tasking */
	
	TaskInit();
	
	/* Now TaskInit() will go to KernelMainLate(), so we don't have anything to do, after all, TaskInit() should never return */
	
	while (1) ;
}
