// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 13:55 BRST
// Last edited on December 18 of 2017, at 19:04 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef __PAGING_H__
#define __PAGING_H__

#include <chicago/multiboot.h>

typedef struct
{
	UInt32 present : 1;
	UInt32 rw : 1;
	UInt32 user : 1;
	UInt32 accessed : 1;
	UInt32 dirty : 1;
	UInt32 unused : 7;
	UInt32 frame : 20;
} Page, *PPage;

typedef struct
{
	Page pages[1024];
} PageTable, *PPageTable;

typedef struct
{
	PPageTable tables[1024];
	UInt32 physical_tables[1024];
	UInt32 physical_address;
} PageDirectory, *PPageDirectory;

UIntPtr PagingRealAlloc(UInt32 size, Boolean align, PUIntPtr phys);
UIntPtr PagingAlloc(UInt32 size);
UIntPtr PagingAlignedAlloc(UInt32 size);
UIntPtr PagingAllocP(UInt32 size, PUIntPtr phys);
UIntPtr PagingAlignedAllocP(UInt32 size, PUIntPtr phys);
Void PagingAllocFrame(PPage page, Int krnl, Int write);
Void PagingFreeFrame(PPage page);
Void PagingEnable(Void);
Void PagingSwitchDirectory(PPageDirectory dir);
PPage PagingGetPage(UInt32 addr, Int make, PPageDirectory dir);
Void PagingCopyPhysicalPage(UInt32 src, UInt32 dst);
PPageTable PagingCloneTable(PPageTable src, PUInt32 phys);
PPageDirectory PagingCloneDirectory(PPageDirectory src);
Void PagingInit(PMultibootHeader mbheader);

#endif		// __PAGING_H__
