// File author is Ítalo Lima Marconato Matias
//
// Created on December 21 of 2017, at 14:40 BRST
// Last edited on December 26 of 2017, at 02:16 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/arch/port.h>
#include <chicago/console.h>
#include <chicago/arch/idt.h>
#include <chicago/keyboard.h>
#include <chicago/task.h>

extern Volatile PProcess ProcessCurrent;

struct		// Keyboard states
{
	UInt32 shift : 1;
	UInt32 alt : 1;
	UInt32 ctrl : 1;
} KeyboardState;

Void KeyboardNormal(Int sc);
Void KeyboardShift(Int sc);
Void KeyboardAlt(Int sc);
Void KeyboardCtrl(Int sc);
Void KeyboardOther(Int sc);

PKeyboardHandler KeyboardUSKeyPressRoutines[] =
{
	0, 0, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal,
	KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal,
	KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal,
	KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardCtrl, KeyboardNormal, KeyboardNormal,
	KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal,
	KeyboardNormal, KeyboardNormal, KeyboardShift, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal,
	KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardNormal, KeyboardShift, KeyboardNormal,
	KeyboardAlt, KeyboardNormal, 0, KeyboardOther, KeyboardOther, KeyboardOther, KeyboardOther, KeyboardOther,
	KeyboardOther, KeyboardOther, KeyboardOther, KeyboardOther, KeyboardOther, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, KeyboardOther,
	KeyboardOther, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
};

Char KeyboardUSLayout[128] =
{
	0, 27,
	'1', '2', '3', '4', '5', '6', '7', '8', '9', '0',
	'-', '=', '\b',
	'\t',
	'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', '\n',
	0,
	'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\'', '`',
	0,
	'\\', 'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/',
	0,
	'*',
	0,
	' ',
	0,
	0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0,
	0,
	0,
	0,
	0,
	0,
	'-',
	0,
	0,
	0,
	'+',
	0,
	0,
	0,
	0,
	0,
	0, 0, 0,
	0,
	0,
	0,
};

Char KeyboardUSLayoutShift[128] =
{
	0, 27,
	'!', '@', '#', '$', '%', '^', '&', '*', '(', ')',
	'_', '+', '\b',
	'\t',
	'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '{', '}', '\n',
	0,
	'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', ':', '"', '~',
	0,
	'|', 'Z', 'X', 'C', 'V', 'B', 'N', 'M', '<', '>', '?',
	0,
	'*',
	0,
	' ',
	0, 
	0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0,
	0,
	0,
	0,
	0,
	0,
	'-',
	0,
	0,
	0,
	'+',
	0,
	0,
	0,
	0,
	0,
	0, 0, 0,
	0,
	0,
	0,
};

PKeyboardKeyHandler KeyboardKeyHandler = Null;
PChar GetStringBuffer = Null;
Int GetStringCollected = 0;
Int GetStringWant = 0;
Boolean GetStringNewLine = False;
Boolean GetStringSilent = False;
Volatile PProcess GetStringProcess = Null;

Void KeyboardHandler(PRegisters regs)
{
	UInt8 scancode = PortInByte(0x60);		// Read the scan code
	PKeyboardHandler handler = KeyboardUSKeyPressRoutines[(Int)scancode & 0x7F];		// Convert to key press routine
	if (scancode == 58)		// It's shift?
		KeyboardShift(58);		// Yes
	else if (handler)		// No, so call the handler if isn't Null
		handler(scancode);
}

Void KeyboardWriteCharacter(Char c)
{
	if (KeyboardKeyHandler)		// Call the key handler if not Null
		KeyboardKeyHandler(c);
}

Void KeyboardNormal(Int s)
{
	if (s == (42 | 0x80) || s == (54 | 0x80))		// Scan code is shift release?
		KeyboardState.shift = 0;		// Yes so unset shift
	else if (s & 0x80)
		return;
	else if (!KeyboardUSLayout[s])		// It's inside our US layout?
		return;		// No...
	else if (KeyboardState.shift)		// Shift?
		KeyboardWriteCharacter(KeyboardUSLayoutShift[s]);		// Yes!
	else if (KeyboardState.ctrl)		// Control?
	{
		Int out = (Int)(KeyboardUSLayoutShift[s] - 0x40);		// Yes!
		if (out < 0 || out > 0x1F)
			KeyboardWriteCharacter(KeyboardUSLayout[s]);
		else
			KeyboardWriteCharacter((Char)out);
	}
	else		// So, it's normal
		KeyboardWriteCharacter(KeyboardUSLayout[s]);
}

Void KeyboardShift(Int s)
{
	KeyboardState.shift ^= 1;
}

Void KeyboardAlt(Int s)
{
	KeyboardState.alt ^= 1;
}

Void KeyboardCtrl(Int s)
{
	KeyboardState.ctrl ^= 1;
}

Void KeyboardOther(Int s)
{ }

Void KeyboardInit(Void)
{
	IDTRegisterIRQHandler(1, KeyboardHandler);
}

Void GetStringHandler(Char c)
{
	Asm Volatile("cli");
	
	if (ProcessCurrent != GetStringProcess)
		Asm Volatile("mov %0, %%cr3" :: "r"(GetStringProcess -> cr3));
	
	if (c == '\b')		// Backspace?
	{
		if (GetStringCollected != 0)		// Yes, so we can decrease?
		{
			if (!GetStringSilent)		// Yes!
				ConsoleWriteFormated("\b \b");
			GetStringBuffer[GetStringCollected] = '\0';
			GetStringCollected--;
		}
		Asm Volatile("mov %0, %%cr3" :: "r"(ProcessCurrent -> cr3));
		Asm Volatile("sti");
		return;
	}
	else if (c == '\n')		// End?
	{
		GetStringNewLine = True;
		if (!GetStringSilent)
			ConsoleWriteFormated("\n");
		Asm Volatile("mov %0, %%cr3" :: "r"(ProcessCurrent -> cr3));
		Asm Volatile("sti");
		return;
	}
	else if (!GetStringSilent)		// Normal or Ctrl+Key
		ConsoleWriteFormated("%c", c);
	if (GetStringCollected < GetStringWant)		// Put in buffer?
	{
		if (ProcessCurrent != GetStringProcess)
			Asm Volatile("mov %0, %%cr3" :: "r"(GetStringProcess -> cr3));
		GetStringBuffer[GetStringCollected] = c;		// Yes!
		GetStringCollected++;
	}
	Asm Volatile("mov %0, %%cr3" :: "r"(ProcessCurrent -> cr3));
	Asm Volatile("sti");
}

Int GetString(PChar buffer, Boolean silent, Int size)
{
	GetStringBuffer = buffer;		// Setup the buffer
	GetStringCollected = 0;		// Zero the collected count
	GetStringWant = size;		// Set the want count
	GetStringNewLine = False;		// Set new line to False
	GetStringSilent = silent;		// Set the silent config
	GetStringProcess = ProcessCurrent;		// Set the caller process
	KeyboardKeyHandler = GetStringHandler;		// Set the key handler
	while ((GetStringCollected < size) && (!GetStringNewLine))		// Just do nothing while collect from the keyboard
		Asm Volatile("hlt");
	KeyboardKeyHandler = Null;		// Remove the key handler
	buffer[GetStringCollected] = '\0';		// Finish the buffer
	return GetStringCollected;		// Return how many character we collected
}
