// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 15:25 BRST
// Last edited on December 21 of 2017, at 12:26 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/arch/paging.h>

extern UIntPtr PagingAllocPointer;
extern PPageDirectory PagingCurrentDirectory;

UIntPtr HeapCurrent = 0;
UIntPtr HeapCurrentAligned = 0;
UIntPtr HeapStart = 0;
Boolean HeapInitialized = False;

UIntPtr HeapGetCurrent(Void)
{
	if (!HeapInitialized)		// Heap initialized?
	{
		HeapCurrent = HeapCurrentAligned = HeapStart = ((PagingAllocPointer + 0x1000) & ~0xFFF);		// No, so let's init it
		PagingAllocPointer = (UIntPtr)-1;		// And DESTROY the PagingAllocPointer, now we can use MemoryAlloc()
		
		PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory);
		PagingAllocFrame(PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory), 0, 1);
		
		Asm Volatile("movl %cr3, %eax; movl %eax, %cr3");
		
		HeapInitialized = True;
	}
	
	return HeapCurrent;
}

UIntPtr HeapGetStart(Void)
{
	if (!HeapInitialized)		// I'm not going to put the same comments in every function...
	{
		HeapCurrent = HeapCurrentAligned = HeapStart = ((PagingAllocPointer + 0x1000) & ~0xFFF);
		
		PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory);
		PagingAllocFrame(PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory), 0, 1);
		
		Asm Volatile("movl %cr3, %eax; movl %eax, %cr3");
		
		HeapInitialized = True;
	}
	
	return HeapStart;
}

UIntPtr HeapIncrement(UIntPtr amount)
{
	if (!HeapInitialized)
	{
		HeapCurrent = HeapCurrentAligned = HeapStart = ((PagingAllocPointer + 0x1000) & ~0xFFF);
		
		PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory);
		PagingAllocFrame(PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory), 0, 1);
		
		Asm Volatile("movl %cr3, %eax; movl %eax, %cr3");
		
		HeapInitialized = True;
	}
	
	if (amount <= 0)
		return -1;
	
	HeapCurrent += amount;
	
	for (; HeapCurrentAligned < HeapCurrent; HeapCurrentAligned += 0x1000)
	{
		PagingGetPage(HeapCurrentAligned, 1, PagingCurrentDirectory);
		PagingAllocFrame(PagingGetPage(HeapCurrentAligned, 1, PagingCurrentDirectory), 0, 1);
	}
	
	Asm Volatile("movl %cr3, %eax; movl %eax, %cr3");
	
	return 0;
}

UIntPtr HeapDecrement(UIntPtr amount)
{
	
	if (!HeapInitialized)
	{
		HeapCurrent = HeapCurrentAligned = HeapStart = ((PagingAllocPointer + 0x1000) & ~0xFFF);
		
		PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory);
		PagingAllocFrame(PagingGetPage(HeapCurrent, 1, PagingCurrentDirectory), 0, 1);
		
		Asm Volatile("movl %cr3, %eax; movl %eax, %cr3");
		
		HeapInitialized = True;
	}
	
	if (amount <= 0)
		return -1;
	else if (HeapCurrent - amount < HeapStart)
		return -1;
		
	HeapCurrent -= amount;
	
	return 0;
}
