// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 13:31 BRST
// Last edited on December 18 of 2017, at 13:31 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/console.h>

extern void GDTLoad(UIntPtr base, UInt16 limit);

UInt8 GDTEntries[5][8];

Void GDTSetGate(UInt8 num, UInt32 base, UInt32 limit, UInt8 type)
{
	if (num >= 5)		// Check num
		Panic("Sorry, CHicago got an fatal error and halted.\nThis fault happened in GDTSetGate() because entry num is invalid.");		// Is invalid, so we need to halt the kernel

	if ((limit > 0xFFFF) && ((limit & 0xFFFF) != 0xFFFF))		// Check limit
		Panic("Sorry, CHicago got an fatal error and halted.\nThis fault happened in GDTSetGate() because limit is invalid.");		// Is invalid, so we need to halt the kernel

	if (limit > 0xFFFF)		// Adjust granularity
	{
		limit = limit >> 12;
		GDTEntries[num][6] = 0xC0;
	}
	else
		GDTEntries[num][6] = 0x40;

	GDTEntries[num][0] = limit & 0xFF;		// Encode the limit
	GDTEntries[num][1] = (limit >> 8) & 0xFF;
	GDTEntries[num][6] |= (limit >> 16) & 0x0F;

	GDTEntries[num][2] = base & 0xFF;		// Encode the base
	GDTEntries[num][3] = (base >> 8) & 0xFF;
	GDTEntries[num][4] = (base >> 16) & 0xFF;
	GDTEntries[num][7] = (base >> 24) & 0xFF;

	GDTEntries[num][5] = type;		// Encode the type
}

Void GDTInit(Void)
{
	GDTSetGate(0, 0, 0, 0);		// Null entry
	GDTSetGate(1, 0, 0xFFFFFFFF, 0x9A);		// Code entry
	GDTSetGate(2, 0, 0xFFFFFFFF, 0x92);		// Data entry
	GDTSetGate(3, 0, 0xFFFFFFFF, 0xFA);		// User mode code entry
	GDTSetGate(4, 0, 0xFFFFFFFF, 0xF2);		// User mode data entry
	GDTLoad((UIntPtr)GDTEntries, 8 * 3 - 1);		// Load new GDT
}
