// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 22:26 BRST
// Last edited on December 28 of 2017, at 21:12 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/console.h>
#include <chicago/string.h>
#include <chicago/arch/idt.h>
#include <chicago/arch/paging.h>
#include <chicago/alloc.h>
#include <chicago/task.h>
#include <chicago/list.h>

// Tasking code from codeOS2

extern PPageDirectory PagingCurrentDirectory;
extern PPageDirectory PagingKernelDirectory;
extern UInt32 PagingAllocPointer;

extern PChar StringDuplicate(PChar src);
extern Void KernelMainLate(Void);

PProcess ProcessCurrent = Null;
PProcessInfo ProcessCurrentInfo = Null;
UInt32 ProcessNextID = 0;
Boolean ProcessEnabled = False;

Void KernelProcess(Void)
{
	/* Finish tasking initialization */
	
	ProcessEnabled = True;
	Message(MESSAGE_LEVEL_INFO, "Tasking initialized\n");
	
	/* Let's return to the kernel entry point */
	
	KernelMainLate();
	
	while (1) ;
}

PProcess ProcessCreate(UInt32 entry)
{	
	PProcess proc = (PProcess)MemoryAlloc(sizeof(Process));
	PUInt32 stack = Null;
	
	if (!proc)
		return Null;
	
	proc -> id = ProcessNextID++;
	proc -> state = 0;
	proc -> executed = False;
	
	proc -> esp = proc -> stack = MemoryAlignedAlloc(0x1000, 0x1000);
	
	if (!proc -> esp)
	{
		MemoryFree((UIntPtr)proc);
		return Null;
	}
	
	stack = (PUInt32)(proc -> esp + 0x1000);
	
	*--stack = 0x202;		// EFlags
	*--stack = 0x08;		// CS
	*--stack = entry;		// EIP
	*--stack = 0;			// EAX
	*--stack = 0;			// EBX
	*--stack = 0;			// ECX
	*--stack = 0;			// EDX
	*--stack = 0;			// ESI
	*--stack = 0;			// EDI
	*--stack = proc -> esp + 0x1000;		// EBP
	*--stack = 0x10;		// DS
	*--stack = 0x10;		// FS
	*--stack = 0x10;		// ES
	*--stack = 0x10;		// GS
	
	proc -> esp = (UInt32)stack;
	proc -> eip = entry;
	proc -> cr3 = PagingCloneDirectory(PagingCurrentDirectory) -> physical_address;
	
	Int reenable = ProcessEnabled;
	
	if (reenable)
		ProcessEnabled = False;
	
	UInt32 oldcr3 = PagingCurrentDirectory -> physical_address;
	
	Asm Volatile("movl %0, %%cr3" :: "r"(proc -> cr3));
	
	proc -> private_data = (PVoid)MemoryAlloc(sizeof(ProcessInfo));
	
	PProcessInfo procinfo = (PProcessInfo)proc -> private_data;
	
	if (!procinfo)
	{
		Asm Volatile("movl %0, %%cr3" :: "r"(oldcr3));
	
		if (reenable)
			ProcessEnabled = True;
		
		MemoryAlignedFree((UIntPtr)proc -> stack);
		MemoryFree((UIntPtr)proc);
		
		return Null;
	}
	
	procinfo -> pid = proc -> id;
	procinfo -> file_list = ListNew();
	
	if (!procinfo -> file_list)
	{
		MemoryFree((UIntPtr)procinfo);
		
		Asm Volatile("movl %0, %%cr3" :: "r"(oldcr3));
	
		if (reenable)
			ProcessEnabled = True;
		
		MemoryAlignedFree((UIntPtr)proc -> stack);
		MemoryFree((UIntPtr)proc);
		
		return Null;
	}
	
	procinfo -> file_count = 3;
	procinfo -> heap_start = procinfo -> heap_current = procinfo -> heap_current_aligned = ((PagingAllocPointer + 0x1000) & ~0xFFF);
	procinfo -> last_error = 0;
	procinfo -> libc_exit_handlers = (PVoid)MemoryAlloc(128);
	procinfo -> libc_exit_handler_ptr = 32;
	procinfo -> libc_file_list = Null;
	
	PagingGetPage(procinfo -> heap_start, 1, PagingCurrentDirectory);
	PagingAllocFrame(PagingGetPage(procinfo -> heap_current, 1, PagingCurrentDirectory), 0, 1);
	
	Asm Volatile("movl %0, %%cr3" :: "r"(oldcr3));
	
	if (reenable)
		ProcessEnabled = True;
	
	return proc;
}

Void ProcessAdd(PProcess proc)
{
	if (!proc)
		return;
	
	ProcessEnabled = False;
	
	proc -> next = ProcessCurrent -> next;
	proc -> next -> prev = ProcessCurrent;
	proc -> prev = ProcessCurrent;
	ProcessCurrent -> next = proc;
	
	ProcessEnabled = True;
}

Void ProcessExit(Void)
{
	ProcessEnabled = False;
	
	PagingSwitchDirectory(PagingKernelDirectory);
	
	if (ProcessCurrent -> id == 0)		// Process is chkrnl?
	{
		ProcessEnabled = False;		// Yes... THIS WASN'T SUPPOSED TO HAPPEN! HALT THE SYSTEM
		Panic("Sorry, CHicago got an fatal error and halted.\nThis fault happened in ProcessExit() because chkrnl can't be killed.");
	}
	
	ProcessEnabled = False;		// No
	
	Asm Volatile("movl %0, %%cr3" :: "r"(ProcessCurrent -> cr3));
	
	ListFree(ProcessCurrentInfo -> file_list);
	MemoryFree((UIntPtr)ProcessCurrentInfo);
	
	PagingSwitchDirectory(PagingKernelDirectory);
	
	MemoryAlignedFree((UIntPtr)ProcessCurrent -> stack);		// TODO: Free the process directory
	MemoryFree((UIntPtr)ProcessCurrent);
	ProcessCurrent -> prev -> next = ProcessCurrent -> next;
	ProcessCurrent -> next -> prev = ProcessCurrent -> prev;
	ProcessCurrent -> state = 2;
	
	ProcessEnabled = True;
	
	ProcessForceSwitch();
}

PProcess ProcessGet(UInt32 pid)
{
	PProcess curr = ProcessCurrent;
	
	do
	{
		if (curr -> id == pid)
			return curr;
		curr = curr -> next;
	}
	while (curr != ProcessCurrent);
	
	return Null;
}

Void ProcessSwitch(Void)
{
	Asm Volatile("push %eax");		// Push current regs
	Asm Volatile("push %ebx");
	Asm Volatile("push %ecx");
	Asm Volatile("push %edx");
	Asm Volatile("push %esi");
	Asm Volatile("push %edi");
	Asm Volatile("push %ebp");
	Asm Volatile("push %ds");
	Asm Volatile("push %es");
	Asm Volatile("push %fs");
	Asm Volatile("push %gs");
	Asm Volatile("mov %%esp, %%eax" : "=a"(ProcessCurrent -> esp));
	
	ProcessCurrent = ProcessCurrent -> next;
	ProcessCurrentInfo = (PProcessInfo)ProcessCurrent -> private_data;
	
	if (!ProcessCurrent -> executed)		// First time running this process?
	{
		ProcessCurrent -> executed = True;		// Yes, so we need to IRET at the end
		
		Asm Volatile("mov %%eax, %%cr3" :: "a"(ProcessCurrent -> cr3));
		Asm Volatile("mov %%eax, %%esp" :: "a"(ProcessCurrent -> esp));
		Asm Volatile("pop %gs");
		Asm Volatile("pop %fs");
		Asm Volatile("pop %es");
		Asm Volatile("pop %ds");
		Asm Volatile("pop %ebp");
		Asm Volatile("pop %edi");
		Asm Volatile("pop %esi");
		Asm Volatile("pop %edx");
		Asm Volatile("pop %ecx");
		Asm Volatile("pop %ebx");
		Asm Volatile("pop %eax");
		Asm Volatile("iret");
		
		return;
	}
	
	Asm Volatile("mov %%eax, %%cr3" :: "a"(ProcessCurrent -> cr3));		// No, so pop everthing and return to ISRHandler32
	Asm Volatile("mov %%eax, %%esp" :: "a"(ProcessCurrent -> esp));
	Asm Volatile("pop %gs");
	Asm Volatile("pop %fs");
	Asm Volatile("pop %es");
	Asm Volatile("pop %ds");
	Asm Volatile("pop %ebp");
	Asm Volatile("pop %edi");
	Asm Volatile("pop %esi");
	Asm Volatile("pop %edx");
	Asm Volatile("pop %ecx");
	Asm Volatile("pop %ebx");
	Asm Volatile("pop %eax");
}

Void ProcessForceSwitch(Void)
{
	if (!ProcessEnabled)
		return;
	Asm Volatile("int $0x3E");
}

Void TaskInit(Void)
{
	IDTSetGate(0x3E, (UInt32)ProcessSwitch , 0x08, 0x8E);		// Register the ProcessForceSwitch() interrupt
	
	ProcessCurrent = ProcessCreate((UInt32)KernelProcess);		// Create kernel process
	ProcessCurrent -> next = ProcessCurrent;
	ProcessCurrent -> prev = ProcessCurrent;
	
	ProcessCurrent -> executed = True;
	
	Asm Volatile("mov %%eax, %%cr3" :: "a"(ProcessCurrent -> cr3));		// Execute kernel process
	Asm Volatile("mov %%eax, %%esp" :: "a"(ProcessCurrent -> esp));
	Asm Volatile("pop %gs");
	Asm Volatile("pop %fs");
	Asm Volatile("pop %es");
	Asm Volatile("pop %ds");
	Asm Volatile("pop %ebp");
	Asm Volatile("pop %edi");
	Asm Volatile("pop %esi");
	Asm Volatile("pop %edx");
	Asm Volatile("pop %ecx");
	Asm Volatile("pop %ebx");
	Asm Volatile("pop %eax");
	Asm Volatile("iret");
	
	while (1) ;
}
