// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 16:09 BRST
// Last edited on December 18 of 2017, at 17:48 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/string.h>
#include <chicago/heap.h>
#include <chicago/alloc.h>

PVoid MemoryAllocBase = Null;

static PMemoryAllocBlock MemoryAllocFindBlock(PMemoryAllocBlock* last, UInt32 size)
{
	PMemoryAllocBlock b = MemoryAllocBase;
	
	while (b && !(b -> free && b -> size >= size))		// Block is free? Block is equal or greater than specified size
	{
		*last = b;		// Yes
		b = b -> next;
	}
	
	return b;
}

static PMemoryAllocBlock MemoryAllocExtendHeap(PMemoryAllocBlock last, UInt32 size)
{
	PMemoryAllocBlock b = (PMemoryAllocBlock)HeapGetCurrent();
	
	if (HeapIncrement(20 + size) == (UIntPtr)-1)		// If it returns -1, something is VERY WRONG
		return Null;	// ... SOMETHING IS WRONG
	
	b -> size = size;
	b -> next = Null;
	b -> prev = last;
	b -> ptr = b -> data;
	
	if (last)
		last -> next = b;
	
	b -> free = False;
	
	return b;
}

static Void MemoryAllocSplitBlock(PMemoryAllocBlock b, UInt32 size)
{
	PMemoryAllocBlock new = (PMemoryAllocBlock)(b -> data + size);
	
	new -> size = b -> size - size - 20;
	new -> next = b -> next;
	new -> prev = b;
	new -> free = True;
	new -> ptr = new -> data;
	
	b -> size = size;
	b -> next = new;
	
	if (new -> next)
		new -> next -> prev = new;
}

static PMemoryAllocBlock MemoryAllocFuseBlock(PMemoryAllocBlock b)
{
	if (b -> next && b -> next -> free)
	{
		b -> size += 20 + b -> next -> size;
		b -> next = b -> next -> next;
		
		if (b -> next)
			b -> next -> prev = b;
	}
	
	return b;
}

static PMemoryAllocBlock MemoryAllocGetBlock(PVoid p)
{
	PChar tmp = p;
	return (p = tmp -= 20);
}

static Boolean MemoryAllocValidAddress(PVoid p)
{
	UIntPtr pp = (UIntPtr)p;
	
	if (MemoryAllocBase)
	{
		if (p > MemoryAllocBase && pp < HeapGetCurrent())
			return (p == (MemoryAllocGetBlock(p)) -> ptr);
	}
	
	return False;
}

static Void MemoryAllocCopyBlock(PMemoryAllocBlock src, PMemoryAllocBlock dst)
{
	PInt sdata = src -> ptr;
	PInt ddata = dst -> ptr;
	
	for (UInt32 i = 0; i * 4 < src -> size && i * 4 < dst -> size; i++)
		ddata[i] = sdata[i];
}

UIntPtr MemoryAlloc(UInt32 size)
{
	if (size == 0)
		return 0;
	
	PMemoryAllocBlock b;
	PMemoryAllocBlock last;
	UInt32 s = ((((size - 1) >> 2) << 2) + 4);
	
	if (MemoryAllocBase)		// First time?
	{
		last = MemoryAllocBase;		// No
		b = MemoryAllocFindBlock(&last, s);		// Try to find a block
		
		if (b)		// We found?
		{
			if ((b -> size - s) >= 16)		// Yes! We can split it?
				MemoryAllocSplitBlock(b, s);		// Yes
			
			b -> free = False;		// NOW THIS BLOCK BELONG TO US
		}
		else
		{
			b = MemoryAllocExtendHeap(last, s);		// No, so let's extend the heap
			
			if (!b)
				return 0;
		}
	}
	else
	{
		b = MemoryAllocExtendHeap(Null, s);		// Yes, so let's init it
		
		if (!b)
			return 0;
		
		MemoryAllocBase = b;
	}
	
	return (UIntPtr)b -> data;
}

UIntPtr MemoryAlignedAlloc(UIntPtr size, UIntPtr align)		// Aligned alloc that i don't remeber from where i got
{
	if (size == 0)
		return 0;
	
	UIntPtr p1;
	PUIntPtr p2;
	Int offset = align - 1 + sizeof(UIntPtr);
	
	if ((p1 = MemoryAlloc(size + offset)) == 0)
		return (UIntPtr)Null;
	
	p2 = (PUIntPtr)((p1 + offset) & ~(align - 1));
	p2[-1] = p1;
	
	return (UIntPtr)p2;
}

Void MemoryFree(UIntPtr block)
{
	PMemoryAllocBlock b;
	
	if (MemoryAllocValidAddress((PVoid)block))		// Valid block?
	{
		b = MemoryAllocGetBlock((PVoid)block);		// Yes!
		b -> free = True;
		
		if (b -> prev && b -> prev -> free)		// We can fuse with previous?
			b = MemoryAllocFuseBlock(b -> prev);		// Yes, so let's do it!
		
		if (b -> next)		// We can fuse with the next?
			MemoryAllocFuseBlock(b);		// Yes, so do it
		else		// No, so...
		{
			if (b -> prev)		// Free the end of the heap
				b -> prev -> next = Null;
			else
				MemoryAllocBase = Null;		// No more block!
			
			UInt32 decr = HeapGetCurrent() - (UIntPtr)b;
			
			HeapDecrement(decr);		// Now, we can do the REAL free
		}
	}
}

Void MemoryAlignedFree(UIntPtr block)		// Aligned free to used with the aligned alloc
{
	MemoryFree(((PUIntPtr)block)[-1]);
}

UIntPtr MemoryCalloc(UIntPtr mul, UInt32 size)
{
	UIntPtr ret = MemoryAlloc(mul * size);
	
	if (ret)
		MemorySet((PUInt8)ret, 0, mul * size);
	
	return ret;
}

UIntPtr MemoryAlignedCalloc(UIntPtr mul, UIntPtr size, UIntPtr align)
{
	UIntPtr ret = MemoryAlignedAlloc(mul * size, align);
	
	if (ret)
		MemorySet((PUInt8)ret, 0, mul * size);
	
	return ret;
}

UIntPtr MemoryRealloc(UIntPtr block, UInt32 size)
{
	UInt32 s;
	PMemoryAllocBlock b;
	PMemoryAllocBlock new;
	UIntPtr newp;
	
	if (!block)
		return MemoryAlloc(size);
	else if (MemoryAllocValidAddress((PVoid)block))
	{
		s = ((((size - 1) >> 2) << 2) + 4);
		b = MemoryAllocGetBlock((PVoid)block);
		
		if (b -> size >= s)
		{
			if (b -> size - s >= 24)
				MemoryAllocSplitBlock(b, s);
		}
		else
		{
			if (b -> next && b -> next -> free && (b -> size + 20 + b -> next -> size) >= s)
			{
				MemoryAllocFuseBlock(b);
				
				if (b -> size - s >= 24)
					MemoryAllocSplitBlock(b, s);
			}
			else
			{
				newp = MemoryAlloc(s);
				
				if (!newp)
					return 0;
				
				new = MemoryAllocGetBlock((PVoid)newp);
				
				MemoryAllocCopyBlock(b, new);
				MemoryFree(block);
				
				return newp;
			}
		}
		
		return block;
	}
	else
		return 0;
}

UIntPtr MemoryAlignedRealloc(UIntPtr block, UInt32 size, UIntPtr align)
{
	UInt32 s;
	PMemoryAllocBlock b;
	PMemoryAllocBlock new;
	UIntPtr newp;
	PVoid realblock = (PVoid)(((PUIntPtr)block)[-1]);
	
	if (!block)
		return MemoryAlignedAlloc(size, align);
	else if (MemoryAllocValidAddress(realblock))
	{
		s = ((((size - 1) >> 2) << 2) + 4);
		b = MemoryAllocGetBlock(realblock);
		
		if (b -> size >= s)
		{
			if (b -> size - s >= 24)
				MemoryAllocSplitBlock(b, s);
		}
		else
		{
			if (b -> next && b -> next -> free && (b -> size + 20 + b -> next -> size) >= s)
			{
				MemoryAllocFuseBlock(b);
				
				if (b -> size - s >= 24)
					MemoryAllocSplitBlock(b, s);
			}
			else
			{
				PVoid newrealblock;
				
				newp = MemoryAlignedAlloc(s, align);
				
				if (!newp)
					return 0;
				
				newrealblock = (PVoid)(((PUIntPtr)newp)[-1]);
				new = MemoryAllocGetBlock(newrealblock);
				
				MemoryAllocCopyBlock(b, new);
				MemoryAlignedFree(block);
				
				return newp;
			}
		}
		
		return block;
	}
	else
		return 0;
}
