// File author is Ítalo Lima Marconato Matias
//
// Created on December 18 of 2017, at 14:37 BRST
// Last edited on December 18 of 2017, at 15:00 BRST
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
// 
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#include <chicago/types.h>

/* ... Simple memory/string copy, set, length, etc */

PUInt8 MemoryCopy(PUInt8 dest, PUInt8 src, Int count)
{
	if (count == 0 || dest == src)
		return dest;
	
	for (Int i = 0; i < count; i++)
		dest[i] = src[i];
	
	return dest;
}

PUInt8 MemorySet(PUInt8 dest, UInt8 val, Int count)
{
	if (!count)
		return dest;
	
	for (Int i = 0; i < count; i++)
		dest[i] = val;
	
	return dest;
}

Int MemoryCompare(Const PVoid m1, Const PVoid m2, Int count)
{
	PUInt8 um1 = (PUInt8)m1;
	PUInt8 um2 = (PUInt8)m2;
	
	while (count-- > 0)
	{
		Int diff;
		if ((diff = (*um1++ - *um2++)) != 0)
			return diff;
	}
	
	return 0;
}


Int StringGetLength(Const PChar str)
{
	Int len = 0;
	for (Int i = 0; str[i]; i++)
		len++;
	return len;
}

Int StringCompare(Const PChar dest, PChar src)
{
	Int i = 0;
	while ((dest[i] == src[i]))
	{
		if (src[i++] == 0)
			return 1;
	}
	return 0;
}

PChar StringCopy(PChar dest, Const PChar src)
{
	return (PChar)MemoryCopy((PUInt8)dest, (PUInt8)src, StringGetLength((PChar)src) + 1);
}

Void StringConcatenate(PChar dest, Const PChar src)
{
	PChar end = dest + StringGetLength(dest);
	MemoryCopy((PUInt8)end, (Const PUInt8)src, StringGetLength((PChar)src));
	end = end + StringGetLength((PChar)src);
	*end = '\0';
}

UInt32 LinearFind(PChar text, Char acc)
{
	UInt32 i = 0;
	while (text[i] != acc)
		i++;
	return (UInt32)(text) + i;
}

UInt32 StringSpan(PChar text, PChar ac)
{
	PChar acc;
	UInt32 size = 0;
	for (PChar ptr = text; *ptr != '\0'; ptr++)
	{
		for (acc = ac; *acc != '\0'; acc++)
		{
			if (*ptr == *acc)
				break;
		}
		if (*acc == '\0')
			return size;
		else
			size++;
	}
	return size;
}

PChar StringScan(PChar text, PChar ac)
{
	while (*text != '\0')
	{
		PChar acc = ac;
		while (*acc != '\0')
		{
			if (*acc++ == *text)
				return (PChar)text;
		}
		text++;
	}
	return 0;
}

PChar StringTokenize(PChar text, PChar delim, PChar* save)
{
	PChar token;
	if (text == 0)
		text = *save;
	text += StringSpan(text, delim);
	if (*text == '\0')
	{
		*save = text;
		return Null;
	}
	token = text;
	text = StringScan(token, delim);
	if (text == 0)
		*save = (PChar)LinearFind(text, '\0');
	else
	{
		*text = '\0';
		*save = text + 1;
	}
	return token;
}
